#!/usr/bin/perl
use warnings;
use strict;
use threads;
use Data::Dumper;
use List::Util qw(sum);
use DBI;

#die "Usage: [cmd] <lambda> <sigma>" unless scalar(@ARGV) == 2;

### result directory
my $res_dir = "out"; system "mkdir $res_dir";

### analysis settings
my $start = 1381708800;
my $end   = 1381750200;

my $act_start = $start - 3600*24*7;

### list of signals (database tables)
my $signals = {
	"os_A"	=>	"event1a_os_tracked_300s",
	"os_B"	=>	"event1b_os_tracked_300s",
	"os_C"	=>	"event1c_os_tracked_300s",
	"os_D"	=>	"event1d_os_tracked_300s",
	"os_E"	=>	"event1e_os_tracked_300s",
	"os_F"	=>	"event1f_os_tracked_300s",
	"os_G"	=>	"event1g_os_tracked_300s",
	"os_H"	=>	"event1h_os_tracked_300s",
	"os_I"	=>	"event1i_os_tracked_300s",

};
my $feature = "query_cnt"; # counter name for diagnostic signals

### trigger signals (database tables)
my $triggers = {
	"querycnt_A"	=>	"event1a_querycnt_bin_300s",
#	"querycnt_B"	=>	"event1b_querycnt_bin_300s",
#	"querycnt_C"	=>	"event1c_querycnt_bin_300s",
#	"querycnt_D"	=>	"event1d_querycnt_bin_300s",
#	"querycnt_E"	=>	"event1e_querycnt_bin_300s",
#	"querycnt_F"	=>	"event1f_querycnt_bin_300s",
#	"querycnt_G"	=>	"event1g_querycnt_bin_300s",
#	"querycnt_H"	=>	"event1h_querycnt_bin_300s",
#	"querycnt_I"	=>	"event1i_querycnt_bin_300s",
	
};
my $trig_feat = "user_cnt_all_all"; # counter name for trigger signals

#my $global_feature_name = "user_cnt_all_all";
my $global_feature_name = "query_cnt";
my $global_variable_name = "os_tracked";

### connect to database
my $dbi = "dbi:Pg:dbname=experiments;host=127.0.0.1;port=5435;";
my $dbh = DBI->connect($dbi,"expusr","xyz")
	or die "Can't Connect to database: $1\n";
	
### query all tables and compute entropies (keep them in memory)
my $H = {}; foreach my $signal (sort keys %$signals) {

	print STDERR "Now computing entropies for signal $signal...\n";
	
	$H->{$signal} = 
		calc_entropy($act_start,$end,$signals->{$signal},$signal,$feature);
}

my $T = {}; foreach my $trigger (sort keys %$triggers) {
	
	print STDERR "Now retrieving timeseries for signal $trigger...\n";
	
	$H->{$trigger} =
		get_querycnt($act_start,$end,$triggers->{$trigger},$trig_feat);
}


my @threads;

foreach my $signal (sort keys %$H) {
	my $th = threads->create('run_ewma_signal',$signal,$H->{$signal});
	push @threads,$th;
}

print "Out of the main loop. Waiting for all threads to finish...\n";
foreach my $th (@threads) {
	$th->join;
}


sub run_ewma_signal {
	my ($name,$ts) = @_;
	for (my $l = 0.01; $l <= 1; $l += 0.01) {
		for(my $s = 0.01; $s <= 0.75; $s += 0.01) {
			$l = sprintf("%.2f", $l); #correct a strange Perl bug
			$s = sprintf("%.2f", $s); #correct a strange Perl bug
			
			&run_ewma($name,$ts,$l,$s);		
		}
	}	
}


### run EWMA algorithm
sub run_ewma {
	my ($name,$ts,$l,$s) = @_;
	
	print STDERR "Now running EWMA($l,$s) on $name...\n";
	
	open OUT, ">$res_dir/ewma_$name\_l$l\_s$s.txt" or die $!;
	
	#my $iter = 60; 
	my $ewma_prev = 0; 
	my $ewma_prev2 = 0;
	
	foreach my $time (sort keys %$ts) {
		
		my $ewma_this = $l * $ts->{$time} + (1 - $l) * $ewma_prev;
		my $pred_this = $l * $ewma_prev   + (1 - $l) * $ewma_prev2;		
		
		if($ewma_prev==0 ) { $ewma_prev  = $ts->{$time} } # first execution
		if($ewma_prev2==0) { $ewma_prev2 = $ts->{$time} } # first execution
		
		if($time >= $start && $time <= $end) {
			
			my $UCL = (1 + $s) * $pred_this; #opt2
			my $LCL = (1 - $s) * $pred_this; #opt2
			#my $UCL = (1 + $s) * $ewma_this;  #opt1
			#my $LCL = (1 - $s) * $ewma_this;  #opt1
			
			my $code = 0;
			if( $ts->{$time} > $UCL ) { $code = 1  }
			if( $ts->{$time} < $LCL ) { $code = -1 }
			
			print OUT "$time\t$ts->{$time}\t$code\n";
		}

		$ewma_prev2 = $ewma_prev;		
		$ewma_prev  = $ewma_this;

		#$iter++;
	}
	
	close OUT;
}


### calculate entropy for a specific signal in a given timebin
sub calc_entropy {
	my ($hstart,$hend,$table,$vname,$fname) = @_;

	my $H = {};
	my $time_stat = $dbh->selectall_hashref(
		#"select serial_time, sum(feature_value), count(distinct $vname) ".
		"select serial_time, sum(feature_value), count(distinct $global_variable_name) ".
		"from $table ".
		"where serial_time between $hstart and $hend and ".
		"  feature_name = '$global_feature_name' ".
		"group by 1",1);
	
	my $list_val = $dbh->selectall_hashref(
		#"select serial_time, $vname, feature_value ".
		"select serial_time, $global_variable_name, feature_value ".
		"from $table ".
		"where serial_time between $hstart and $hend and ".
		"  feature_name = '$global_feature_name' ", [1,2]);
	
	foreach my $time (sort keys %$list_val) {
		my $var_sum = $time_stat->{$time}->{sum};
		my $var_num = $time_stat->{$time}->{count};
		my $h = 0;
		foreach my $elem (keys %{$list_val->{$time}}) {
			my $val = $list_val->{$time}->{$elem}->{feature_value};
			if($elem eq "") { $var_num++; $elem = "NULL" }
			$h-=($val/$var_sum)*log($val/$var_sum);
		}
		$h = $h/log($var_num);
		$H->{$time} = $h;
		
		if($time >= $start && $time <= $end) {
			print "$time\t$h\t$vname\n";
		}
	}
	
	return $H;
}

### 
sub get_querycnt {
	my ($start,$end,$table,$fname) = @_;

	my $TS = {};
	my $res = $dbh->selectall_hashref(
		"select serial_time, sum(feature_value) ".
		"from $table ".
		"where serial_time between $start and $end and ".
		#"  feature_name = '$fname' ".
		"  feature_name = '$global_feature_name' ".
		"group by 1 order by 1",1);
	
	#print "tab=$table, $start,$end\n"; exit;
		
	my $res2; foreach my $time (sort keys %$res) {
		$res2->{$time} = $res->{$time}->{sum};
	}
	
	return $res2;
}


###

